/* gselt: GTK selection tool. GPL v2 or above; azz@gnu.org */

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <regex.h>
#include <unistd.h>

GtkWidget *window;
guchar *last = NULL;

struct _entry {
  GtkWidget *button;
  char *label;
  char *command;
  regex_t re;
};
typedef struct _entry entry;

#define MAXTYPES 100
entry types[MAXTYPES];
int ntypes = 0;
int under_mouse = 0;

#define BUFSIZE 1024

char *get_match(char *buf, regmatch_t *match) {
  int l = match->rm_eo - match->rm_so;
  char *ret = malloc(l + 1);
  
  strncpy(ret, buf + match->rm_so, l);
  ret[l] = '\0';

  return ret;
}

void read_config(char *name) {
  FILE *f = fopen(name, "r");
  char buf[BUFSIZE], elem[BUFSIZE];
  regex_t re;
  regmatch_t match[4];

  if (!f) {
    fprintf(stderr, "Unable to read config file\n");
    exit(20);
  }

  regcomp(&re, "^(.*)!(.*)!(.*)$", REG_EXTENDED);

  while (fgets(buf, BUFSIZE-1, f) && ntypes < MAXTYPES) {
    int rc;

    buf[strlen(buf) - 1] = '\0';
    rc = regexec(&re, buf, 4, match, 0);
    if (rc == REG_NOMATCH
	|| match[1].rm_so < 0 || match[2].rm_so < 0 || match[3].rm_so < 0) {
      fprintf(stderr, "Unrecognised line: %s\n", buf);
    } else {
      char *ms = get_match(buf, &match[1]);

      regcomp(&types[ntypes].re, ms, REG_EXTENDED | REG_ICASE | REG_NOSUB);
      free(ms);
      types[ntypes].label = get_match(buf, &match[2]);
      types[ntypes].command = get_match(buf, &match[3]);
      
      ++ntypes;
    }
  }

  fclose(f);
}

int timer_function(void *data) { 
  gtk_selection_convert(window, GDK_SELECTION_PRIMARY,
			GDK_TARGET_STRING, GDK_CURRENT_TIME);
  return 1;
}

void handle_button_select(GtkWidget *button, gpointer data) {
  gchar buf[BUFSIZE];

  snprintf(buf, BUFSIZE-1, (gchar *)data, last);
  system(buf);
}

void handle_selection(GtkWidget *window, GdkEvent *event) {
  guchar *buf, *this;
  gint length, format, i, any = 0;
  GdkAtom atom;

  length = gdk_selection_property_get(window->window, &buf, &atom,
				      &format);

  if (length) {
    if (!last || strcmp(last, buf)) {
      if (last) g_free(last);
      last = buf;
    } else {
      return;
    }
  } else {
    if (!last) return;
    last = NULL;
  }

  gtk_widget_hide(window);

  for (i=0; i<ntypes; i++) {
    if (last && regexec(&types[i].re, last, 0, NULL, 0) != REG_NOMATCH) {
      any = 1;
      gtk_widget_show(types[i].button);
    } else {
      gtk_widget_hide(types[i].button);
    }
  }

  if (any) {
    gint x, y, window_w, window_h, screen_w, screen_h;

    /* Get the window size. Yes, this is the wrong order, but GTK
       doesn't calculate the new window size until the window is
       shown. */
    gtk_widget_show(window);
    gdk_window_get_geometry(NULL, NULL, NULL,
			    &screen_w, &screen_h, NULL);
    gdk_window_get_geometry(window->window, NULL, NULL,
			    &window_w, &window_h, NULL);

    if (under_mouse) {
      /* Place the window under the mouse pointer. */
      gint window_x, window_y;
      gdk_window_get_origin(window->window, &window_x, &window_y);
      gdk_window_get_pointer(window->window, &x, &y, NULL);
      x += window_x;
      y += window_y;
    } else {
      /* Place the window in the top right corner. */
      x = screen_w - window_w;
      y = 0;
    }

    /* Make sure the window fits on the screen. */
    if (x < 0) x = 0;
    if (x > (screen_w - window_w)) x = screen_w - window_w;
    if (y < 0) y = 0;
    if (y > (screen_h - window_h)) y = screen_h - window_h;

    gdk_window_move(window->window, x, y);
  }
}

int main(int argc, char **argv) {
  GtkWidget *vbox;
  gint timer, i;
  char buf[BUFSIZE];

  gtk_init(&argc, &argv);
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

  while (1) {
    int c = getopt(argc, argv, "m");
    if (c == -1) break;

    switch (c) {
    case 'm':
      under_mouse = 1;
      break;
    default:
      fprintf(stderr, "Usage: gselt [-m]\n");
      return 20;
    }
  }

  snprintf(buf, BUFSIZE - 1, "%s/.gselt", getenv("HOME"));
  read_config(buf);

  gtk_window_set_title(GTK_WINDOW(window), "gselt");
  gtk_window_set_policy(GTK_WINDOW(window), TRUE, TRUE, TRUE);
  gtk_widget_realize(window);
  gdk_window_set_override_redirect(window->window, TRUE);
  gtk_signal_connect(GTK_OBJECT(window), "selection_received",
		     GTK_SIGNAL_FUNC(handle_selection), NULL);

  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), vbox);

  for (i=0; i<ntypes; i++) {
    types[i].button = gtk_button_new_with_label(types[i].label);
    gtk_box_pack_start(GTK_BOX(vbox), types[i].button, TRUE, TRUE, 0);
    gtk_signal_connect(GTK_OBJECT(types[i].button), "clicked",
		       GTK_SIGNAL_FUNC(handle_button_select), 
		       types[i].command);
  }

  gtk_widget_show(vbox);
  
  timer = gtk_timeout_add(100, timer_function, NULL);

  gtk_main();

  exit(0);
}

